//
// Pretty printer for the Java grammar.
// Author: Kevin Tao, taokr@cs
//
// VERSION: 2c
//
// Newly added:
//    2c  - 9/21/97: Modified for use with JTB 0.9.0b; visitXXX() methods
//          changed to visit() to use overloading
//    2b  - Minor cosmetic changes, changed visibility of "spc" to protected.
//    2a  - Corrected errors in WhileStatement, and Break and ContinueStatement.
//    2a  - Corrected a few slight formatting errors.
//    2   - flushWriter() method to flush the output stream.
//
// (reminders for myself):
// - spc.spc should be printed after every "\n" or println().
// - println() should not be the last thing printed in a visit method.
//

/* RCS Information
 *
 * $Id: PrettyPrinter.java,v 1.2 1997/10/25 20:41:21 ktao Exp $
 *
 * $Log: PrettyPrinter.java,v $
 * Revision 1.2  1997/10/25 20:41:21  ktao
 * - Added log info.
 * - Deleted irrelevant comment.
 *
 */

package visitor;

//import misc.Spacing;
import generated.*;
import java.util.*;
import java.io.*;

public class PrettyPrinter extends Visitor {
    protected int anonymousClassLevel = 0;
    
    protected PrintWriter out;
    protected Spacing spc = new Spacing(4);   // see below for class Spacing
    protected PrintStack printer;

    protected static final Separator newLine = new Separator("", 0, true);
    protected static final Separator doubleNewLine =
	new Separator("\n", 0, true);
    protected static final Separator emptySeparator = new Separator("", 0);
    protected static final Separator primarySpace = new Separator(" ", 1);

    public static final int MAX_PRIORITY = 16;
    protected static final Separator[] spaces = new Separator[30];

    int parenthesesLevel = 0;

    protected void updateParenthesesLevel(int increment) {
	parenthesesLevel += MAX_PRIORITY * increment;
    }

    public int getParenthesesLevel() {
	return parenthesesLevel;
    }
    
    protected Separator getSpace(int priority) {
	if (spaces[priority] != null)
	    return spaces[priority];
	else {
	    spaces[priority] = new Separator(" ", priority);
	    return spaces[priority];
	}
    }

    public PrettyPrinter(PrintWriter o) {
	out = o; 
	printer = new PrintStack(this, out, spc, 80);
    }
    
    public PrettyPrinter() { this(new PrintWriter(System.out, true)); }
    public PrettyPrinter(Writer o)        { this(new PrintWriter(o, true)); }
    public PrettyPrinter(OutputStream o)  { this(new PrintWriter(o, true)); }

    public void flushWriter()     { out.flush(); }

    //
    // A few convenience methods overloading visit()
    //

    public void visit(Enumeration e, Separator innerSeparator) {
	while (e.hasMoreElements()) {
	    ((Node)e.nextElement()).accept(this);
	    if ( e.hasMoreElements() )
		printer.push(innerSeparator);
	}
    }

    public void visit(Enumeration e, Separator innerSeparator,
		      Separator sequenceSeparator) {
	while (e.hasMoreElements()) {
	    visit((NodeSequence)e.nextElement(), sequenceSeparator);
	    if ( e.hasMoreElements() )
		printer.push(innerSeparator);
	}
    }
	
    public void visit(NodeList n, Separator innerSeparator,
		      Separator finalSeparator) {
	visit(n.elements(), innerSeparator);
	printer.push(finalSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator,
		      Separator finalSeparator) {
	if (n.present()) {
	    visit(n.elements(), innerSeparator);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeList n, Separator innerSeparator) {
	visit(n.elements(), innerSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator) {
	if (n.present())
	    visit(n.elements(), innerSeparator);
    }
    
    public void visit(NodeList n, Separator innerSeparator,
		      Separator finalSeparator, Separator sequenceSeparator) {
	visit(n.elements(), innerSeparator, sequenceSeparator);
	printer.push(finalSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator,
		      Separator finalSeparator, Separator sequenceSeparator) {
	if (n.present()) {
	    visit(n.elements(), innerSeparator, sequenceSeparator);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeOptional n, Separator finalSeparator) {
	if ( n.present() ) {
	    n.node.accept(this);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeOptional n, Separator finalSeparator,
		      Separator sequenceSeparator) {
	if ( n.present() ) {
	    visit((NodeSequence) n.node, sequenceSeparator);
	    printer.push(finalSeparator);
	}
    }

    //
    // For convenience, trusts that the node passed to it is a NodeSequence.
    // (of course, will throw an exception if it isn't).
    //
    public void visit(NodeSequence n, Separator sequenceSeparator) {
	visit(n.elements(), sequenceSeparator);
    }

    
    public void visit(Token n) {
	String text = null;
	for (Enumeration e = n.getSpecialTokens(); e.hasMoreElements();) {
	    text = ((Token) e.nextElement()).toString();
	    printer.push(text);
	    printer.push(newLine);
	}
	printer.push(n.toString());
    }

    
    //
    // Auto class visitors--probably don't need to be overridden.
    //

    //
    // User-generated visitor methods below
    //


    protected void updateAnonymousClassLevel(int increment) {
	anonymousClassLevel += increment;
    }
    
    
    protected static final Separator tripleNewLine =
	new Separator("\n \n", 0, true);
    //
    // f0 -> [ PackageDeclaration() ]
    // f1 -> ( ImportDeclaration() )*
    // f2 -> ( TypeDeclaration() )*
    // f3 -> < EOF >
    //
    public void visit(CompilationUnit n) {
	visit(n.f0, doubleNewLine);
	visit(n.f1, newLine, doubleNewLine);
	visit(n.f2, tripleNewLine, newLine);
    }

    //
    // f0 -> "package"
    // f1 -> Name()
    // f2 -> ";"
    //
    public void visit(PackageDeclaration n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "import"
    // f1 -> Name()
    // f2 -> [ "." "*" ]
    // f3 -> ";"
    //
    // uses the toString() method of the Tokens to print manually
    // (rather than visiting the tokens).
    //
    public void visit(ImportDeclaration n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
    }

    //
    // f0 -> ClassDeclaration()
    //       | InterfaceDeclaration()
    //       | ";"
    //
    public void visit(TypeDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ( "abstract"| "final" | "public"  )*
    // f1 -> UnmodifiedClassDeclaration()
    //
    public void visit(ClassDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
    }

    //
    // f0 -> "class"
    // f1 -> < IDENTIFIER >
    // f2 -> [ "extends" Name() ]
    // f3 -> [ "implements" NameList() ]
    // f4 -> ClassBody()
    //
    public void visit(UnmodifiedClassDeclaration n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2, primarySpace, primarySpace);
	visit(n.f3, primarySpace, primarySpace);
	visit(n.f4);
    }

    //
    // f0 -> "{"
    // f1 -> ( ClassBodyDeclaration() )*
    // f2 -> "}"
    //
    public void visit(ClassBody n) {
	visit(n.f0);
	printer.push(newLine);
	if (anonymousClassLevel == 0)
	    printer.push(newLine);
       
	if ( n.f1.present() ) {
	    spc.visitorUpdate(+1);
	    visit(n.f1, (anonymousClassLevel == 0 ? doubleNewLine : newLine),
		  newLine);
	    spc.visitorUpdate(-1);
	}
      
	if (anonymousClassLevel == 0)
	    printer.push(newLine);

	visit(n.f2);
    }

    //
    // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
    // f1 -> UnmodifiedClassDeclaration()
    //
    public void visit(NestedClassDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
    }

    //
    // f0 -> Initializer()
    //       | NestedClassDeclaration()
    //       | NestedInterfaceDeclaration()
    //       | ConstructorDeclaration()
    //       | MethodDeclaration()
    //       | FieldDeclaration()
    //
    public void visit(ClassBodyDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" |
    //              "final" | "native" | "synchronized"  )*
    // f1 -> ResultType()
    // f2 -> < IDENTIFIER >
    // f3 -> "("
    //
    // For lookahead only, so don't need to do this
    public void visit(MethodDeclarationLookahead n) {}

    //
    // f0 -> ( "abstract"| "public"  )*
    // f1 -> UnmodifiedInterfaceDeclaration()
    //
    public void visit(InterfaceDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
    }

    //
    // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
    // f1 -> UnmodifiedInterfaceDeclaration()
    //
    public void visit(NestedInterfaceDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
    }

    //
    // f0 -> "interface"
    // f1 -> < IDENTIFIER >
    // f2 -> [ "extends" NameList() ]
    // f3 -> "{"
    // f4 -> ( InterfaceMemberDeclaration() )*
    // f5 -> "}"
    //
    public void visit(UnmodifiedInterfaceDeclaration n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2, primarySpace, primarySpace);
       
	visit(n.f3);
	printer.push(newLine);
	if (anonymousClassLevel == 0)
	    printer.push(newLine);
       
	if ( n.f4.present() ) {
	    spc.visitorUpdate(+1);
	    visit(n.f4, (anonymousClassLevel == 0 ? doubleNewLine : newLine),
		  newLine);
	    spc.visitorUpdate(-1);
	}
      
	if (anonymousClassLevel == 0)
	    printer.push(newLine);

	visit(n.f5);
    }

    //
    // f0 -> NestedClassDeclaration()
    //       | NestedInterfaceDeclaration()
    //       | MethodDeclaration()
    //       | FieldDeclaration()
    //
    public void visit(InterfaceMemberDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ("public"|"protected"|"private"|"static"|"final"|"transient"|
    //          "volatile")*
    // f1 -> Type()
    // f2 -> VariableDeclarator()
    // f3 -> ( "," VariableDeclarator() )*
    // f4 -> ";"
    //
    public void visit(FieldDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
	visit(n.f3, emptySeparator, emptySeparator, primarySpace);
	visit(n.f4);
    }

    //
    // f0 -> VariableDeclaratorId()
    // f1 -> [ "=" VariableInitializer() ]
    //
    public void visit(VariableDeclarator n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, emptySeparator, primarySpace);
	}
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ( "[" "]" )*
    //
    public void visit(VariableDeclaratorId n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> ArrayInitializer()
    //       | Expression()
    //
    public void visit(VariableInitializer n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "{"
    // f1 -> [ VariableInitializer() ( "," VariableInitializer() )* ]
    // f2 -> [ "," ]
    // f3 -> "}"
    //
    public void visit(ArrayInitializer n) {
	visit(n.f0);
	if (n.f1.present()) {
	    NodeSequence seq = (NodeSequence) n.f1.node;
	    seq.elementAt(0).accept(this);
	    visit((NodeListOptional) seq.elementAt(1), emptySeparator,
		  emptySeparator, primarySpace);
	}
	visit(n.f2, primarySpace);
	visit(n.f3);
    }

    //
    // f0 -> ( "public"| "protected" | "private" | "static" | "abstract"
    //           | "final" | "native" | "synchronized"  )*
    // f1 -> ResultType()
    // f2 -> MethodDeclarator()
    // f3 -> [ "throws" NameList() ]
    // f4 -> ( Block()| ";"  )
    //
    public void visit(MethodDeclaration n) {
	visit(n.f0, primarySpace, primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
	if ( n.f3.present() ) {
	    printer.push(primarySpace);
	    visit(n.f3, emptySeparator, primarySpace);
	}
	printer.push(primarySpace); 
	n.f4.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> FormalParameters()
    // f2 -> ( "[" "]" )*
    //
    public void visit(MethodDeclarator n) {
	visit(n.f0);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "("
    // f1 -> [ FormalParameter() ( "," FormalParameter() )* ]
    // f2 -> ")"
    //
    public void visit(FormalParameters n) {
	visit(n.f0);
	if (n.f1.present()) {
	    NodeSequence seq = (NodeSequence) n.f1.node;
	    seq.elementAt(0).accept(this);
	    visit((NodeListOptional) seq.elementAt(1), emptySeparator,
		  emptySeparator, primarySpace);
	}
	visit(n.f2);
    }

    //
    // f0 -> [ "final" ]
    // f1 -> Type()
    // f2 -> VariableDeclaratorId()
    //
    public void visit(FormalParameter n) {
	visit(n.f0, primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
    }

    //
    // f0 -> [ "public"| "protected" | "private"  ]
    // f1 -> < IDENTIFIER >
    // f2 -> FormalParameters()
    // f3 -> [ "throws" NameList() ]
    // f4 -> "{"
    // f5 -> [ ExplicitConstructorInvocation() ]
    // f6 -> ( BlockStatement() )*
    // f7 -> "}"
    //
    public void visit(ConstructorDeclaration n) {
	visit(n.f0, primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
	printer.push(primarySpace);
	visit(n.f3, primarySpace, primarySpace);
	visit(n.f4);
      
	if (n.f5.present() || n.f6.present()) {
	    printer.push(newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f5, newLine);
	    visit(n.f6, newLine, emptySeparator);
	    spc.visitorUpdate(-1);
	    printer.push(newLine);
	}

	visit(n.f7);
    }

    //
    // f0 -> "this" Arguments() ";"
    //       | [ PrimaryExpression() "." ] "super" Arguments() ";"
    //
    public void visit(ExplicitConstructorInvocation n) {
	if (n.f0.which == 0) {
	    n.f0.choice.accept(this);
	} else {
	    NodeSequence seq = (NodeSequence)n.f0.choice;
	    if (((NodeOptional) seq.elementAt(0)).present()) {
		seq.elementAt(0).accept(this);
		printer.push(primarySpace);
	    }
	    seq.elementAt(1).accept(this);
	    seq.elementAt(2).accept(this);
	    seq.elementAt(3).accept(this);
	}
    }

    //
    // f0 -> [ "static" ]
    // f1 -> Block()
    //
    public void visit(Initializer n) {
	visit(n.f0, primarySpace);
	visit(n.f1);
    }

    //
    // f0 -> ( PrimitiveType()| Name()  )
    // f1 -> ( "[" "]" )*
    //
    public void visit(Type n) {
	n.f0.choice.accept(this);
	visit(n.f1);
    }

    //
    // f0 -> "boolean"
    //       | "char"
    //       | "byte"
    //       | "short"
    //       | "int"
    //       | "long"
    //       | "float"
    //       | "double"
    //
    public void visit(PrimitiveType n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "void"
    //       | Type()
    //
    public void visit(ResultType n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ( "." < IDENTIFIER >  )*
    //
    public void visit(Name n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> Name()
    // f1 -> ( "," Name() )*
    //
    public void visit(NameList n) {
	visit(n.f0);
	visit(n.f1, emptySeparator, emptySeparator, primarySpace);
    }

    //
    // f0 -> ConditionalExpression()
    // f1 -> [ AssignmentOperator() Expression() ]
    //
    public void visit(Expression n) {
	updateParenthesesLevel(+1);
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(getSpace(13));
	    visit(n.f1, emptySeparator, getSpace(14));
	}
	updateParenthesesLevel(-1);
    }

    //
    // f0 -> "="
    //       | "*="
    //       | "/="
    //       | "%="
    //       | "+="
    //       | "-="
    //       | "<<="
    //       | ">>="
    //       | ">>>="
    //       | "&="
    //       | "^="
    //       | "|="
    //
    public void visit(AssignmentOperator n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ConditionalOrExpression()
    // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
    //
    public void visit(ConditionalExpression n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(primarySpace);
	    visit(n.f1, emptySeparator, getSpace(13));
	}
    }

    //
    // f0 -> ConditionalAndExpression()
    // f1 -> ( "||" ConditionalAndExpression() )*
    //
    public void visit(ConditionalOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(12), emptySeparator, getSpace(11));
	}
    }

    //
    // f0 -> InclusiveOrExpression()
    // f1 -> ( "&&" InclusiveOrExpression() )*
    //
    public void visit(ConditionalAndExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(11), emptySeparator, getSpace(10));
	}
    }

    //
    // f0 -> ExclusiveOrExpression()
    // f1 -> ( "|" ExclusiveOrExpression() )*
    //
    public void visit(InclusiveOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(10), emptySeparator, getSpace(9));
	}
    }

    //
    // f0 -> AndExpression()
    // f1 -> ( "^" AndExpression() )*
    //
    public void visit(ExclusiveOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(9), emptySeparator, getSpace(8));
	}
    }

    //
    // f0 -> EqualityExpression()
    // f1 -> ( "&" EqualityExpression() )*
    //
    public void visit(AndExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(8), emptySeparator, getSpace(7));
	}
    }

    //
    // f0 -> InstanceOfExpression()
    // f1 -> ( ( "=="| "!="  ) InstanceOfExpression() )*
    //
    public void visit(EqualityExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(7), emptySeparator, getSpace(6));
	}
    }

    //
    // f0 -> RelationalExpression()
    // f1 -> [ "instanceof" Type() ]
    //
    public void visit(InstanceOfExpression n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(primarySpace);
	    visit(n.f1, emptySeparator, getSpace(2));
	}
    }

    //
    // f0 -> ShiftExpression()
    // f1 -> ( ( "<"| ">" | "<=" | ">="  ) ShiftExpression() )*
    //
    public void visit(RelationalExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(6), emptySeparator, getSpace(5));
	}
    }

    //
    // f0 -> AdditiveExpression()
    // f1 -> ( ( "<<"| ">>" | ">>>"  ) AdditiveExpression() )*
    //
    public void visit(ShiftExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(5), emptySeparator, getSpace(4));
	}
    }

    //
    // f0 -> MultiplicativeExpression()
    // f1 -> ( ( "+"| "-"  ) MultiplicativeExpression() )*
    //
    public void visit(AdditiveExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(4), emptySeparator, getSpace(3));
	}
    }

    //
    // f0 -> UnaryExpression()
    // f1 -> ( ( "*"| "/" | "%"  ) UnaryExpression() )*
    //
    public void visit(MultiplicativeExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1, getSpace(3), emptySeparator, getSpace(2));
	}
    }

    //
    // f0 -> ( "+"| "-"  ) UnaryExpression()
    //       | PreIncrementExpression()
    //       | PreDecrementExpression()
    //       | UnaryExpressionNotPlusMinus()
    //
    public void visit(UnaryExpression n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "++"
    // f1 -> PrimaryExpression()
    //
    public void visit(PreIncrementExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> "--"
    // f1 -> PrimaryExpression()
    //
    public void visit(PreDecrementExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> ( "~"| "!"  ) UnaryExpression()
    //       | CastExpression()
    //       | PostfixExpression()
    //
    public void visit(UnaryExpressionNotPlusMinus n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "(" PrimitiveType()
    //       | "(" Name() "[" "]"
    //       | "(" Name() ")" ( "~" | "!" | "(" | < IDENTIFIER >  | "this"
    //                                       | "super" | "new" | Literal()  )
    //
    public void visit(CastLookahead n) {}

    //
    // f0 -> PrimaryExpression()
    // f1 -> [ "++"| "--"  ]
    //
    public void visit(PostfixExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> "(" Type() ")" UnaryExpression()
    //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
    //
    public void visit(CastExpression n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	for (int i = 0; i < 3; ++i) 
	    seq.elementAt(i).accept(this);
	printer.push(getSpace(2));
	seq.elementAt(3).accept(this);
    }

    //
    // f0 -> PrimaryPrefix()
    // f1 -> ( PrimarySuffix() )*
    //
    public void visit(PrimaryExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
   // f0 -> Literal()
   //       | "this"
   //       | "super" "." <IDENTIFIER>
   //       | "(" Expression() ")"
   //       | AllocationExpression()
   //       | ResultType() "." "class"
   //       | Name()
    //
    public void visit(PrimaryPrefix n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "." "this"
    //       | "." AllocationExpression()
    //       | "[" Expression() "]"
    //       | "." < IDENTIFIER >
    //       | Arguments()
    //
    public void visit(PrimarySuffix n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < INTEGER_LITERAL >
    //       | < FLOATING_POINT_LITERAL >
    //       | < CHARACTER_LITERAL >
    //       | < STRING_LITERAL >
    //       | BooleanLiteral()
    //       | NullLiteral()
    //
    public void visit(Literal n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "true"
    //       | "false"
    //
    public void visit(BooleanLiteral n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "null"
    //
    public void visit(NullLiteral n) {
	visit(n.f0);
    }


    public static final Separator openingParentheses = new Separator("", 2);
    //
    // f0 -> "("
    // f1 -> [ ArgumentList() ]
    // f2 -> ")"
    //
    public void visit(Arguments n) {
	visit(n.f0);
	printer.push(openingParentheses);
	visit(n.f1);
	visit(n.f2);
    }


    //
    // f0 -> Expression()
    // f1 -> ( "," Expression() )*
    //
    public void visit(ArgumentList n) {
	visit(n.f0);
	visit(n.f1, emptySeparator, emptySeparator, getSpace(2));
    }

    //
    // f0 -> "new" PrimitiveType() ArrayDimsAndInits()
    //       | "new" Name() ( ArrayDimsAndInits()
    //                        | Arguments() [ ClassBody() ] )
    //
    public void visit(AllocationExpression n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;

	seq.elementAt(0).accept(this);
	printer.push(getSpace(2));
	seq.elementAt(1).accept(this);
	if (n.f0.which == 0)
	    seq.elementAt(2).accept(this);
	else {
	    NodeChoice nodeChoice = (NodeChoice) seq.elementAt(2);
	    if (nodeChoice.which == 0)
		nodeChoice.choice.accept(this);
	    else {
		seq = (NodeSequence) nodeChoice.choice;
		seq.elementAt(0).accept(this);
		if (((NodeOptional) seq.elementAt(1)).present()) {
		    printer.push(primarySpace);
		    updateAnonymousClassLevel(+1);
		    seq.elementAt(1).accept(this);
		    updateAnonymousClassLevel(-1);
		}
	    }
	}
    }

    
    //
    // f0 -> ( "[" Expression() "]" )+ ( "[" "]" )*
    //       | ( "[" "]" )+ ArrayInitializer()
    //
    public void visit(ArrayDimsAndInits n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	if (n.f0.which == 0)
	    seq.accept(this);
	else {
	    seq.elementAt(0).accept(this);
	    printer.push(primarySpace);
	    seq.elementAt(1).accept(this);
	}
    }


    //
    // f0 -> LabeledStatement()
    //       | Block()
    //       | EmptyStatement()
    //       | StatementExpression() ";"
    //       | SwitchStatement()
    //       | IfStatement()
    //       | WhileStatement()
    //       | DoStatement()
    //       | ForStatement()
    //       | BreakStatement()
    //       | ContinueStatement()
    //       | ReturnStatement()
    //       | ThrowStatement()
    //       | SynchronizedStatement()
    //       | TryStatement()
    //
    public void visit(Statement n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ":"
    // f2 -> Statement()
    //
    public void visit(LabeledStatement n) {
	visit(n.f0);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
    }

    //
    // f0 -> "{"
    // f1 -> ( BlockStatement() )*
    // f2 -> "}"
    //
    public void visit(Block n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f1, newLine, newLine);
	    spc.visitorUpdate(-1);
	}
	visit(n.f2);
    }

    //
    // f0 -> LocalVariableDeclaration() ";"
    //       | Statement()
    //       | UnmodifiedClassDeclaration()
    //       | UnmodifiedInterfaceDeclaration()
   //
    public void visit(BlockStatement n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> [ "final" ]
    // f1 -> Type()
    // f2 -> VariableDeclarator()
    // f3 -> ( "," VariableDeclarator() )*
    //
    public void visit(LocalVariableDeclaration n) {
	visit(n.f0, primarySpace);
	visit(n.f1);
	printer.push(primarySpace);
	visit(n.f2);
	visit(n.f3, emptySeparator, emptySeparator, primarySpace);
    }

    //
    // f0 -> ";"
    //
    public void visit(EmptyStatement n) {
	visit(n.f0);
    }

    //
    // f0 -> PreIncrementExpression()
    //       | PreDecrementExpression()
    //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
    //
    public void visit(StatementExpression n) {
	if (n.f0.which == 2) {
	    NodeSequence seq = (NodeSequence) n.f0.choice;
	    NodeOptional nodeOptional = (NodeOptional) seq.elementAt(1);
	    if (nodeOptional.present()
		   && ((NodeChoice) nodeOptional.node).which == 2) {
		seq.elementAt(0).accept(this);
		printer.push(primarySpace);
		visit((NodeSequence) ((NodeChoice) nodeOptional.node).choice,
		      primarySpace);
		return;
	    }
	}
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "switch"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> "{"
    // f5 -> ( SwitchLabel() ( BlockStatement() )* )*
    // f6 -> "}"
    //
    public void visit(SwitchStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
	printer.push(primarySpace);
	visit(n.f4);
	printer.push(newLine);
	spc.visitorUpdate(+1);
	for (Enumeration e = n.f5.elements(); e.hasMoreElements();) {
	    NodeSequence seq = (NodeSequence) e.nextElement();
	    seq.elementAt(0).accept(this);
	    printer.push(primarySpace);
	    NodeListOptional list = (NodeListOptional) seq.elementAt(1);
	    if (list.present()) {
		printer.push(newLine);
		spc.visitorUpdate(+1);
		visit(list, newLine, newLine);
		
		if (e.hasMoreElements()) {
		    NodeChoice nodeChoice = (NodeChoice)
			((BlockStatement) list.elementAt(list.size() - 1)).f0;
		    if (nodeChoice.which == 1) {
			switch (((NodeChoice)
				 ((Statement) nodeChoice.choice).f0) .which) {
			case 9: case 10: case 11: break;
			default:
			    printer.push("// fall through");
			    printer.push(newLine);
			}
		    }
		}
		spc.visitorUpdate(-1);
	    } else if (!e.hasMoreElements()) {
		printer.push(newLine);
	    }
	}
	spc.visitorUpdate(-1);
	visit(n.f6);
    }

    //
    // f0 -> "case" Expression() ":"
    //       | "default" ":"
    //
    public void visit(SwitchLabel n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	seq.elementAt(0).accept(this);
	if (n.f0.which == 0) {
	    printer.push(primarySpace);
	    seq.elementAt(1).accept(this);
	    seq.elementAt(2).accept(this);
	} else
	    seq.elementAt(1).accept(this);
    }

    //
    // f0 -> "if"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Statement()
    // f5 -> [ "else" Statement() ]
    //
    public void visit(IfStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);

	boolean isIfBlock = (n.f4.f0.which == 1);
	if (!isIfBlock) {    // only indent if it's not a Block
	    spc.visitorUpdate(+1);
	    printer.push(newLine);
	} else
	    printer.push(primarySpace);
	visit(n.f4);
	if ( n.f4.f0.which != 1 )
	    spc.visitorUpdate(-1);

	if ( n.f5.present() ) {
	    if (!isIfBlock)
		printer.push(newLine);
	    else 
		printer.push(primarySpace);

	    int kindOfElseBlock = ((Statement)
				   ((NodeSequence) n.f5.node).elementAt(1))
		                  .f0.which;
	  
	    if (kindOfElseBlock != 1 && kindOfElseBlock != 5) {
		NodeSequence seq = (NodeSequence) n.f5.node;
		seq.elementAt(0).accept(this);
		printer.push(newLine);
		spc.visitorUpdate(+1);
		seq.elementAt(1).accept(this);
		spc.visitorUpdate(-1);
	    } else
		visit((NodeSequence) n.f5.node, primarySpace);
	}
    }

    //
    // f0 -> "while"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Statement()
    //
    public void visit(WhileStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);

	if (n.f4.f0.which != 1) {
	    printer.push(newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f4);
	    spc.visitorUpdate(-1);
	} else {
	    printer.push(primarySpace);
	    visit(n.f4);
	}
    }

    //
    // f0 -> "do"
    // f1 -> Statement()
    // f2 -> "while"
    // f3 -> "("
    // f4 -> Expression()
    // f5 -> ")"
    // f6 -> ";"
    //
    public void visit(DoStatement n) {
	visit(n.f0);

	if (n.f1.f0.which != 1) {
	    printer.push(newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f1);
	    spc.visitorUpdate(-1);
	    printer.push(newLine);
	} else {
	    printer.push(primarySpace);
	    visit(n.f1);
	    printer.push(primarySpace);
	}
	visit(n.f2);
	printer.push(primarySpace);
	visit(n.f3);
	visit(n.f4);
	visit(n.f5);
	visit(n.f6);
    }

    //
    // f0 -> "for"
    // f1 -> "("
    // f2 -> [ ForInit() ]
    // f3 -> ";"
    // f4 -> [ Expression() ]
    // f5 -> ";"
    // f6 -> [ ForUpdate() ]
    // f7 -> ")"
    // f8 -> Statement()
    //
    public void visit(ForStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
	if (n.f4.present()) {
	    printer.push(primarySpace);
	    visit(n.f4);
	}
	visit(n.f5);
	if (n.f6.present()) {
	    printer.push(primarySpace);
	    visit(n.f6);
	}
	visit(n.f7);
      
	if ( n.f8.f0.which != 1 ) {
	    printer.push(newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f8);
	    spc.visitorUpdate(-1);
	} else {
	    printer.push(primarySpace);
	    visit(n.f8); 
	}
    }

    //
    // f0 -> LocalVariableDeclaration()
    //       | StatementExpressionList()
    //
    public void visit(ForInit n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> StatementExpression()
    // f1 -> ( "," StatementExpression() )*
    //
    public void visit(StatementExpressionList n) {
	visit(n.f0);
	visit(n.f1, emptySeparator, emptySeparator, primarySpace);
    }

    //
    // f0 -> StatementExpressionList()
    //
    public void visit(ForUpdate n) {
	visit(n.f0);
    }

    //
    // f0 -> "break"
    // f1 -> [ < IDENTIFIER >  ]
    // f2 -> ";"
    //
    public void visit(BreakStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "continue"
    // f1 -> [ < IDENTIFIER >  ]
    // f2 -> ";"
    //
    public void visit(ContinueStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "return"
    // f1 -> [ Expression() ]
    // f2 -> ";"
    //
    public void visit(ReturnStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(primarySpace);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "throw"
    // f1 -> Expression()
    // f2 -> ";"
    //
    public void visit(ThrowStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "synchronized"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Block()
    //
    public void visit(SynchronizedStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
	printer.push(primarySpace);
	visit(n.f4);
    }

    //
    // f0 -> "try"
    // f1 -> Block()
    // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
    // f3 -> [ "finally" Block() ]
    //
    public void visit(TryStatement n) {
	visit(n.f0);
	printer.push(primarySpace);
	visit(n.f1);
	for ( Enumeration e = n.f2.elements(); e.hasMoreElements(); ) {
	    NodeSequence seq = (NodeSequence) e.nextElement();
	    printer.push(primarySpace);
	    seq.elementAt(0).accept(this);
	    printer.push(primarySpace);
	    seq.elementAt(1).accept(this);
	    seq.elementAt(2).accept(this);
	    seq.elementAt(3).accept(this);
	    printer.push(primarySpace);
	    seq.elementAt(4).accept(this);
	}

	if (n.f3.present()) {
	    printer.push(primarySpace);
	    visit(n.f3, emptySeparator, primarySpace);
	}
    }
}

//
// For indentation
//
class Spacing {
    final int INDENT_AMT;

    String spc = "";
    int indentLevel = 0;
    int deepness = 0;
    int preparedUpdate = 0;
    
    public Spacing(int indentAmt) {
	INDENT_AMT = indentAmt;
    }

    public String toString() {
	return spc;
    }

    public int getDeepness() {
	return deepness;
    }

    public void visitorUpdate(int numIndentLvls) {
	preparedUpdate += numIndentLvls;
    }

    public void notifyVisitorUpdates() {
	if (preparedUpdate != 0) {
	    printerUpdate(preparedUpdate);
	    preparedUpdate = 0;
	}
    }
    
    public void printerUpdate(int numIndentLvls) {
	indentLevel += numIndentLvls;
        int deepnessIncrement = numIndentLvls * INDENT_AMT;
	deepness += deepnessIncrement;
	//	System.out.println("printerUpdate: #######" + indentLevel + " " + deepness);
	if (numIndentLvls < 0) {
	    spc = spc.substring(-deepnessIncrement);
	} else if (numIndentLvls > 0) {
	    StringBuffer buf = new StringBuffer(spc);
	    for (int i = 0; i < deepnessIncrement; ++i)
		buf.append(" ");
	    spc = buf.toString();
	}
    }
}



class PrintStack {

    final PrintWriter out;
    final int lineLength;
    final Spacing spc;
    final PrettyPrinter visitor;
    int currentColumn;
    
    
    boolean lineBroken = false;
	
    public PrintStack(PrettyPrinter visitor, PrintWriter out, Spacing spc,
		      int lineLength) {
	this.visitor = visitor;
	this.out = out;
	this.spc = spc;
	this.lineLength = lineLength;
    }

    protected Vector items = new Vector();
    
    int offset = 0;
    int columnOffset = 0;
    
    protected void flush() {
	int length = items.size();
	out.print(spc.spc);
	for (int i = offset; i < length; ++i) {
	    out.print(items.elementAt(i).toString());
	}
	out.flush();
	offset = 0;
	columnOffset = 0;
	items.removeAllElements();
	if (lineBroken) {
	    spc.printerUpdate(-2);
	    lineBroken = false;
	}
    }

    protected void flushPart(int upto) {
	out.print(spc.spc);
	for (int i = offset; i < upto; ++i) {
	    out.print(items.elementAt(i).toString());
	}
	out.println(((Item) items.elementAt(upto)).output);
	offset = upto + 1;
	if (!lineBroken) {
	    spc.printerUpdate(+2);
	    lineBroken = true;
	}
    }
    

    protected static class Item {
	public String output;
	public Separator separator;
	public int endColumn;
	public int priority;
	public Item(String output, Separator separator, int endColumn,
		    int parenthesesLevel) {
	    this.output = output;
	    this.separator = separator;
	    priority = separator.priority - parenthesesLevel;
	    this.endColumn = endColumn;
	}
	public String toString() {
	    return output + separator;
	}
    }


    protected int findMaximalSeparator() {
	int length = items.size();
	int maxPriority = -100000;
	int maxPriorityIndex = 0; // default (-> only one item
	// which is too long)
	for (int i = offset; i < length; ++i) {
	    Item currentItem = (Item) items.elementAt(i);
	    if (!lineBroken && currentItem.endColumn <= spc.getDeepness()
		+ 2 * spc.INDENT_AMT)
		continue;
	    // >= because we want to find the last maximal one
	    int priority = currentItem.priority;
	    //System.out.println("item#" + currentItem.toString() + "# has priority " + priority);
	    if (priority >= maxPriority) {
		maxPriority = priority;
		maxPriorityIndex = i;
	    }
	}
	return maxPriorityIndex;
    }


    protected String waitingOutput = null;

    protected final Separator emptySeparator = new Separator("", 0);
						   
    public void push(String output) {
	if (waitingOutput != null)
	    push(emptySeparator);
		 
	waitingOutput = output;
    }

    
    public void push(Separator newSeparator) {
	if (waitingOutput == null)
	    waitingOutput = "";
	    
	if (items.size() == 0) {
	    spc.notifyVisitorUpdates();
	    currentColumn = spc.getDeepness();
	}
	
	if (newSeparator.isLineBreak()) {
	    items.addElement(new Item(waitingOutput, newSeparator,
				      currentColumn,
				      visitor.getParenthesesLevel()));
	    flush();
	    out.println();
	    spc.notifyVisitorUpdates();
	} else {
	    int increaseInLength = waitingOutput.length() +
		newSeparator.length();
	    currentColumn += increaseInLength;

	    if (currentColumn - columnOffset >= lineLength - 1) {
		if (items.size() == 0) {
		    items.addElement(new Item(waitingOutput, newSeparator,
					      currentColumn,
					      visitor.getParenthesesLevel()));
		    /*		    System.out.println("@@@@@@@@@@@@@@@@@@" + waitingOutput + newSeparator + "@@@@" + currentColumn + " " + spc.getDeepness() + " " + increaseInLength + " " + columnOffset);*/
		    
		    flushPart(0);
		    //#####columnoffset
		} else {
		    /* System.out.println("==================================================");
                  System.out.println("#" + waitingOutput + newSeparator.toString() + "# couldn't be added, because: " + increaseInLength + " " + currentColumn + " " + columnOffset + " " + offset);
		    System.out.println("It would look like: ");
		    
		    out.print("##" + spc.spc);
	for (int i = offset; i < items.size(); ++i) {
	    out.print(items.elementAt(i).toString());
	}
	out.println("##");
	System.out.println("--------------------------------------------------------------------------");*/
	
		    int maximalSeparator = findMaximalSeparator();
		    flushPart(maximalSeparator);
		    columnOffset =
			((Item) items.elementAt(maximalSeparator)).endColumn -
			spc.getDeepness();
		    
		    items.addElement(new Item(waitingOutput, newSeparator,
					      currentColumn,
					      visitor.getParenthesesLevel()));
		}
	    } else {
		items.addElement(new Item(waitingOutput, newSeparator,
					  currentColumn,
					  visitor.getParenthesesLevel()));
	    }
	}
	waitingOutput = null;
    }

}


class Separator {

    protected boolean lineBreak;
    protected String text;
    protected int priority;
    
    public boolean isLineBreak() { return lineBreak; }
    public int getPriority() { return priority; }
    public String toString() { return text; }
    public int length() { return text.length(); }
    
    public Separator(String text, int priority) {
	this(text, priority, false);
    }

    public Separator(String text, int priority, boolean lineBreak) {
	this.text = text;
	this.priority = priority;
	this.lineBreak = lineBreak;
    }

}

