//
// Pretty printer for the Java grammar.
// Author: Kevin Tao, taokr@cs
//
// VERSION: 2c
//
// Newly added:
//    2c  - 9/21/97: Modified for use with JTB 0.9.0b; visitXXX() methods
//          changed to visit() to use overloading
//    2b  - Minor cosmetic changes, changed visibility of "spc" to protected.
//    2a  - Corrected errors in WhileStatement, and Break and ContinueStatement.
//    2a  - Corrected a few slight formatting errors.
//    2   - flushWriter() method to flush the output stream.
//
// (reminders for myself):
// - spc.spc should be printed after every "\n" or println().
// - println() should not be the last thing printed in a visit method.
//

/* RCS Information
 *
 * $Id: PrettyPrinter.java,v 1.2 1997/10/25 20:41:21 ktao Exp $
 *
 * $Log: PrettyPrinter.java,v $
 * Revision 1.2  1997/10/25 20:41:21  ktao
 * - Added log info.
 * - Deleted irrelevant comment.
 *
 */

package visitor;

import generated.*;
import java.util.*;
import java.io.*;

public class PrettyPrinter extends Visitor {
    protected PrintWriter out;
    protected Spacing spc = new Spacing(4);   // see below for class Spacing
    protected PrintStack printer;

    public PrettyPrinter(PrintWriter o) {
	out = o; 
	printer = new PrintStack(out, spc, 80);
    }
    
    public PrettyPrinter() { this(new PrintWriter(System.out, true)); }
    public PrettyPrinter(Writer o)        { this(new PrintWriter(o, true)); }
    public PrettyPrinter(OutputStream o)  { this(new PrintWriter(o, true)); }

    public void flushWriter()     { out.flush(); }

    //
    // A few convenience methods overloading visit()
    //

    public void visit(Enumeration e, Separator innerSeparator) {
	while (e.hasMoreElements()) {
	    ((Node)e.nextElement()).accept(this);
	    if ( e.hasMoreElements() )
		printer.push(innerSeparator);
	}
    }

    public void visit(Enumeration e, Separator innerSeparator,
		      Separator sequenceSeparator) {
	while (e.hasMoreElements()) {
	    visit((NodeSequence)e.nextElement(), sequenceSeparator);
	    if ( e.hasMoreElements() )
		printer.push(innerSeparator);
	}
    }
	
    public void visit(NodeList n, Separator innerSeparator,
		      Separator finalSeparator) {
	visit(n.elements(), innerSeparator);
	printer.push(finalSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator,
		      Separator finalSeparator) {
	if (n.present()) {
	    visit(n.elements(), innerSeparator);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeList n, Separator innerSeparator) {
	visit(n.elements(), innerSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator) {
	if (n.present())
	    visit(n.elements(), innerSeparator);
    }
    
    public void visit(NodeList n, Separator innerSeparator,
		      Separator finalSeparator, Separator sequenceSeparator) {
	visit(n.elements(), innerSeparator, sequenceSeparator);
	printer.push(finalSeparator);
    }

    public void visit(NodeListOptional n, Separator innerSeparator,
		      Separator finalSeparator, Separator sequenceSeparator) {
	if (n.present()) {
	    visit(n.elements(), innerSeparator, sequenceSeparator);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeOptional n, Separator finalSeparator) {
	if ( n.present() ) {
	    n.node.accept(this);
	    printer.push(finalSeparator);
	}
    }
    
    public void visit(NodeOptional n, Separator finalSeparator,
		      Separator sequenceSeparator) {
	if ( n.present() ) {
	    visit((NodeSequence) n.node, sequenceSeparator);
	    printer.push(finalSeparator);
	}
    }

    //
    // For convenience, trusts that the node passed to it is a NodeSequence.
    // (of course, will throw an exception if it isn't).
    //
    public void visit(NodeSequence n, Separator sequenceSeparator) {
	visit(n.elements(), sequenceSeparator);
    }

    
    public void visit(Token n) {
	for (Enumeration e = new PrecedingSpecialTokensEnumeration(n);
	     e.hasMoreElements();) 
	    printer.pushSpecialToken((Token) e.nextElement());
	printer.push(n.toString());
	for (Enumeration e = new FollowingSpecialTokensEnumeration(n);
	     e.hasMoreElements();) 
	    printer.pushSpecialToken((Token) e.nextElement());
    }

    
    //
    // Auto class visitors--probably don't need to be overridden.
    //

    //
    // User-generated visitor methods below
    //

    protected int anonymousClassLevel = 0;
    protected void updateAnonymousClassLevel(int increment) {
	anonymousClassLevel += increment;
    }
    
    
    //
    // f0 -> [ PackageDeclaration() ]
    // f1 -> ( ImportDeclaration() )*
    // f2 -> ( TypeDeclaration() )*
    // f3 -> < EOF >
    //
    public void visit(CompilationUnit n) {
	visit(n.f0, Separators.doubleNewLine);
	visit(n.f1, Separators.newLine, Separators.doubleNewLine);
	visit(n.f2, Separators.tripleNewLine, Separators.newLine);
    }

    //
    // f0 -> "package"
    // f1 -> Name()
    // f2 -> ";"
    //
    public void visit(PackageDeclaration n) {
	visit(n.f0);
	printer.push(Separators.declarator);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "import"
    // f1 -> Name()
    // f2 -> [ "." "*" ]
    // f3 -> ";"
    //
    // uses the toString() method of the Tokens to print manually
    // (rather than visiting the tokens).
    //
    public void visit(ImportDeclaration n) {
	visit(n.f0);
	printer.push(Separators.declarator);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
    }

    //
    // f0 -> ClassDeclaration()
    //       | InterfaceDeclaration()
    //       | ";"
    //
    public void visit(TypeDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ( "abstract"| "final" | "public"  )*
    // f1 -> UnmodifiedClassDeclaration()
    //
    public void visit(ClassDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
    }

    //
    // f0 -> "class"
    // f1 -> < IDENTIFIER >
    // f2 -> [ "extends" Name() ]
    // f3 -> [ "implements" NameList() ]
    // f4 -> ClassBody()
    //
    public void visit(UnmodifiedClassDeclaration n) {
	visit(n.f0);
	printer.push(Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2, Separators.declarator, Separators.subDeclarator);
	visit(n.f3, Separators.declarator, Separators.subDeclarator);
	visit(n.f4);
    }

    //
    // f0 -> "{"
    // f1 -> ( ClassBodyDeclaration() )*
    // f2 -> "}"
    //
    public void visit(ClassBody n) {
	visit(n.f0);
	printer.push(Separators.newLine);
	if (anonymousClassLevel == 0)
	    printer.push(Separators.newLine);
       
	if ( n.f1.present() ) {
	    spc.visitorUpdate(+1);
	    visit(n.f1,
		  (anonymousClassLevel == 0 ? Separators.doubleNewLine
		   : Separators.newLine),
		  Separators.newLine);
	    spc.visitorUpdate(-1);
	}
      
	if (anonymousClassLevel == 0)
	    printer.push(Separators.newLine);

	visit(n.f2);
    }

    //
    // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
    // f1 -> UnmodifiedClassDeclaration()
    //
    public void visit(NestedClassDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
    }

    //
    // f0 -> Initializer()
    //       | NestedClassDeclaration()
    //       | NestedInterfaceDeclaration()
    //       | ConstructorDeclaration()
    //       | MethodDeclaration()
    //       | FieldDeclaration()
    //
    public void visit(ClassBodyDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ( "public"| "protected" | "private" | "static" | "abstract" |
    //              "final" | "native" | "synchronized"  )*
    // f1 -> ResultType()
    // f2 -> < IDENTIFIER >
    // f3 -> "("
    //
    // For lookahead only, so don't need to do this
    public void visit(MethodDeclarationLookahead n) {}

    //
    // f0 -> ( "abstract"| "public"  )*
    // f1 -> UnmodifiedInterfaceDeclaration()
    //
    public void visit(InterfaceDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
    }

    //
    // f0 -> ("static"|"abstract"|"final"|"public"|"protected"|"private")*
    // f1 -> UnmodifiedInterfaceDeclaration()
    //
    public void visit(NestedInterfaceDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
    }

    //
    // f0 -> "interface"
    // f1 -> < IDENTIFIER >
    // f2 -> [ "extends" NameList() ]
    // f3 -> "{"
    // f4 -> ( InterfaceMemberDeclaration() )*
    // f5 -> "}"
    //
    public void visit(UnmodifiedInterfaceDeclaration n) {
	visit(n.f0);
	printer.push(Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2, Separators.declarator, Separators.subDeclarator);
       
	visit(n.f3);
	printer.push(Separators.newLine);
	if (anonymousClassLevel == 0)
	    printer.push(Separators.newLine);
       
	if ( n.f4.present() ) {
	    spc.visitorUpdate(+1);
	    visit(n.f4,
		  (anonymousClassLevel == 0 ? Separators.doubleNewLine
		   : Separators.newLine),
		  Separators.newLine);
	    spc.visitorUpdate(-1);
	}
      
	if (anonymousClassLevel == 0)
	    printer.push(Separators.newLine);

	visit(n.f5);
    }

    //
    // f0 -> NestedClassDeclaration()
    //       | NestedInterfaceDeclaration()
    //       | MethodDeclaration()
    //       | FieldDeclaration()
    //
    public void visit(InterfaceMemberDeclaration n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ("public"|"protected"|"private"|"static"|"final"|"transient"|
    //          "volatile")*
    // f1 -> Type()
    // f2 -> VariableDeclarator()
    // f3 -> ( "," VariableDeclarator() )*
    // f4 -> ";"
    //
    public void visit(FieldDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2);
	visit(n.f3, Separators.empty, Separators.empty, Separators.declarator);
	visit(n.f4);
    }

    //
    // f0 -> VariableDeclaratorId()
    // f1 -> [ "=" VariableInitializer() ]
    //
    public void visit(VariableDeclarator n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeAssignment);
	    visit(n.f1, Separators.empty, Separators.afterAssignment);
	}
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ( "[" "]" )*
    //
    public void visit(VariableDeclaratorId n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> ArrayInitializer()
    //       | Expression()
    //
    public void visit(VariableInitializer n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "{"
    // f1 -> [ VariableInitializer() ( "," VariableInitializer() )* ]
    // f2 -> [ "," ]
    // f3 -> "}"
    //
    public void visit(ArrayInitializer n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.openingArguments);
	    NodeSequence seq = (NodeSequence) n.f1.node;
	    seq.elementAt(0).accept(this);
	    visit((NodeListOptional) seq.elementAt(1), Separators.empty,
		  Separators.empty, Separators.betweenArguments);
	}
	visit(n.f2, Separators.betweenArguments);
	visit(n.f3);
	if (n.f1.present())
	    printer.push(Separators.closingArguments);
    }

    //
    // f0 -> ( "public"| "protected" | "private" | "static" | "abstract"
    //           | "final" | "native" | "synchronized"  )*
    // f1 -> ResultType()
    // f2 -> MethodDeclarator()
    // f3 -> [ "throws" NameList() ]
    // f4 -> ( Block()| ";"  )
    //
    public void visit(MethodDeclaration n) {
	visit(n.f0, Separators.declarator, Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2);
	if ( n.f3.present() ) {
	    printer.push(Separators.declarator);
	    visit(n.f3, Separators.empty, Separators.subDeclarator);
	}
	printer.push(Separators.declarator); 
	n.f4.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> FormalParameters()
    // f2 -> ( "[" "]" )*
    //
    public void visit(MethodDeclarator n) {
	visit(n.f0);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "("
    // f1 -> [ FormalParameter() ( "," FormalParameter() )* ]
    // f2 -> ")"
    //
    public void visit(FormalParameters n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.declaratorEmpty);
	    NodeSequence seq = (NodeSequence) n.f1.node;
	    seq.elementAt(0).accept(this);
	    visit((NodeListOptional) seq.elementAt(1), Separators.empty,
		  Separators.empty, Separators.declarator);
	}
	visit(n.f2);
    }

    //
    // f0 -> [ "final" ]
    // f1 -> Type()
    // f2 -> VariableDeclaratorId()
    //
    public void visit(FormalParameter n) {
	visit(n.f0, Separators.subDeclarator);
	visit(n.f1);
	printer.push(Separators.subDeclarator);
	visit(n.f2);
    }

    //
    // f0 -> [ "public"| "protected" | "private"  ]
    // f1 -> < IDENTIFIER >
    // f2 -> FormalParameters()
    // f3 -> [ "throws" NameList() ]
    // f4 -> "{"
    // f5 -> [ ExplicitConstructorInvocation() ]
    // f6 -> ( BlockStatement() )*
    // f7 -> "}"
    //
    public void visit(ConstructorDeclaration n) {
	visit(n.f0, Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2);
	printer.push(Separators.declarator);
	visit(n.f3, Separators.declarator, Separators.subDeclarator);
	visit(n.f4);
      
	if (n.f5.present() || n.f6.present()) {
	    printer.push(Separators.newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f5, Separators.newLine);
	    visit(n.f6, Separators.newLine, Separators.newLine);
	    spc.visitorUpdate(-1);
	}

	visit(n.f7);
    }

    //
    // f0 -> "this" Arguments() ";"
    //       | [ PrimaryExpression() "." ] "super" Arguments() ";"
    //
    public void visit(ExplicitConstructorInvocation n) {
	if (n.f0.which == 0) {
	    n.f0.choice.accept(this);
	} else {
	    NodeSequence seq = (NodeSequence)n.f0.choice;
	    if (((NodeOptional) seq.elementAt(0)).present()) {
		seq.elementAt(0).accept(this);
		printer.push(Separators.subDeclarator);
	    }
	    seq.elementAt(1).accept(this);
	    seq.elementAt(2).accept(this);
	    seq.elementAt(3).accept(this);
	}
    }

    //
    // f0 -> [ "static" ]
    // f1 -> Block()
    //
    public void visit(Initializer n) {
	visit(n.f0, Separators.control);
	visit(n.f1);
    }

    //
    // f0 -> ( PrimitiveType()| Name()  )
    // f1 -> ( "[" "]" )*
    //
    public void visit(Type n) {
	n.f0.choice.accept(this);
	visit(n.f1);
    }

    //
    // f0 -> "boolean"
    //       | "char"
    //       | "byte"
    //       | "short"
    //       | "int"
    //       | "long"
    //       | "float"
    //       | "double"
    //
    public void visit(PrimitiveType n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "void"
    //       | Type()
    //
    public void visit(ResultType n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ( "." < IDENTIFIER >  )*
    //
    public void visit(Name n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> Name()
    // f1 -> ( "," Name() )*
    //
    public void visit(NameList n) {
	visit(n.f0);
	visit(n.f1, Separators.empty, Separators.empty,
	      Separators.subDeclarator);
    }

    //
    // f0 -> ConditionalExpression()
    // f1 -> [ AssignmentOperator() Expression() ]
    //
    public void visit(Expression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeAssignment);
	    visit(n.f1, Separators.empty, Separators.afterAssignment);
	}
    }

    //
    // f0 -> "="
    //       | "*="
    //       | "/="
    //       | "%="
    //       | "+="
    //       | "-="
    //       | "<<="
    //       | ">>="
    //       | ">>>="
    //       | "&="
    //       | "^="
    //       | "|="
    //
    public void visit(AssignmentOperator n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> ConditionalOrExpression()
    // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
    //
    public void visit(ConditionalExpression n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(Separators.conditional);
	    visit(n.f1, Separators.empty, Separators.conditional);
	}
    }

    //
    // f0 -> ConditionalAndExpression()
    // f1 -> ( "||" ConditionalAndExpression() )*
    //
    public void visit(ConditionalOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeConditionalAnd);
	    visit(n.f1, Separators.afterConditionalAnd, Separators.empty,
		  Separators.beforeConditionalAnd);
	}
    }

    //
    // f0 -> InclusiveOrExpression()
    // f1 -> ( "&&" InclusiveOrExpression() )*
    //
    public void visit(ConditionalAndExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeInclusiveOr);
	    visit(n.f1, Separators.afterInclusiveOr, Separators.empty,
		  Separators.beforeInclusiveOr);
	}
    }

    //
    // f0 -> ExclusiveOrExpression()
    // f1 -> ( "|" ExclusiveOrExpression() )*
    //
    public void visit(InclusiveOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeExclusiveOr);
	    visit(n.f1, Separators.afterExclusiveOr, Separators.empty,
		  Separators.beforeExclusiveOr);
	}
    }

    //
    // f0 -> AndExpression()
    // f1 -> ( "^" AndExpression() )*
    //
    public void visit(ExclusiveOrExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeExclusiveOr);
	    visit(n.f1, Separators.afterExclusiveOr, Separators.empty,
		  Separators.beforeExclusiveOr);
	}
    }

    //
    // f0 -> EqualityExpression()
    // f1 -> ( "&" EqualityExpression() )*
    //
    public void visit(AndExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeAnd);
	    visit(n.f1, Separators.afterAnd, Separators.empty,
		  Separators.beforeAnd);
	}
    }

    //
    // f0 -> InstanceOfExpression()
    // f1 -> ( ( "=="| "!="  ) InstanceOfExpression() )*
    //
    public void visit(EqualityExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeEquality);
	    visit(n.f1, Separators.afterEquality, Separators.empty,
		  Separators.beforeEquality);
	}
    }

    //
    // f0 -> RelationalExpression()
    // f1 -> [ "instanceof" Type() ]
    //
    public void visit(InstanceOfExpression n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(Separators.beforeInstanceOf);
	    visit(n.f1, Separators.empty, Separators.afterInstanceOf);
	}
    }

    //
    // f0 -> ShiftExpression()
    // f1 -> ( ( "<"| ">" | "<=" | ">="  ) ShiftExpression() )*
    //
    public void visit(RelationalExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeRelational);
	    visit(n.f1, Separators.afterRelational, Separators.empty,
		  Separators.beforeRelational);
	}
    }

    //
    // f0 -> AdditiveExpression()
    // f1 -> ( ( "<<"| ">>" | ">>>"  ) AdditiveExpression() )*
    //
    public void visit(ShiftExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeShift);
	    visit(n.f1, Separators.afterShift, Separators.empty,
		  Separators.beforeShift);
	}
    }

    //
    // f0 -> MultiplicativeExpression()
    // f1 -> ( ( "+"| "-"  ) MultiplicativeExpression() )*
    //
    public void visit(AdditiveExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeAdditive);
	    visit(n.f1, Separators.afterAdditive, Separators.empty,
		  Separators.beforeAdditive);
	}
    }

    //
    // f0 -> UnaryExpression()
    // f1 -> ( ( "*"| "/" | "%"  ) UnaryExpression() )*
    //
    public void visit(MultiplicativeExpression n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.beforeMultiplicative);
	    visit(n.f1, Separators.afterMultiplicative, Separators.empty,
		  Separators.beforeMultiplicative);
	}
    }

    //
    // f0 -> ( "+"| "-"  ) UnaryExpression()
    //       | PreIncrementExpression()
    //       | PreDecrementExpression()
    //       | UnaryExpressionNotPlusMinus()
    //
    public void visit(UnaryExpression n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "++"
    // f1 -> PrimaryExpression()
    //
    public void visit(PreIncrementExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> "--"
    // f1 -> PrimaryExpression()
    //
    public void visit(PreDecrementExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> ( "~"| "!"  ) UnaryExpression()
    //       | CastExpression()
    //       | PostfixExpression()
    //
    public void visit(UnaryExpressionNotPlusMinus n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "(" PrimitiveType()
    //       | "(" Name() "[" "]"
    //       | "(" Name() ")" ( "~" | "!" | "(" | < IDENTIFIER >  | "this"
    //                                       | "super" | "new" | Literal()  )
    //
    public void visit(CastLookahead n) {}

    //
    // f0 -> PrimaryExpression()
    // f1 -> [ "++"| "--"  ]
    //
    public void visit(PostfixExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
    // f0 -> "(" Type() ")" UnaryExpression()
    //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
    //
    public void visit(CastExpression n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	for (int i = 0; i < 3; ++i) 
	    seq.elementAt(i).accept(this);
	printer.push(Separators.cast);
	seq.elementAt(3).accept(this);
    }

    //
    // f0 -> PrimaryPrefix()
    // f1 -> ( PrimarySuffix() )*
    //
    public void visit(PrimaryExpression n) {
	visit(n.f0);
	visit(n.f1);
    }

    //
   // f0 -> Literal()
   //       | "this"
   //       | "super" "." <IDENTIFIER>
   //       | "(" Expression() ")"
   //       | AllocationExpression()
   //       | ResultType() "." "class"
   //       | Name()
    //
    public void visit(PrimaryPrefix n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "." "this"
    //       | "." AllocationExpression()
    //       | "[" Expression() "]"
    //       | "." < IDENTIFIER >
    //       | Arguments()
    //
    public void visit(PrimarySuffix n) {
	if (n.f0.which == 3)
	    printer.push(Separators.beforeDot);
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < INTEGER_LITERAL >
    //       | < FLOATING_POINT_LITERAL >
    //       | < CHARACTER_LITERAL >
    //       | < STRING_LITERAL >
    //       | BooleanLiteral()
    //       | NullLiteral()
    //
    public void visit(Literal n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "true"
    //       | "false"
    //
    public void visit(BooleanLiteral n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "null"
    //
    public void visit(NullLiteral n) {
	visit(n.f0);
    }


    //
    // f0 -> "("
    // f1 -> [ ArgumentList() ]
    // f2 -> ")"
    //
    public void visit(Arguments n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.openingArguments);
	    visit(n.f1);
	    visit(n.f2);
	    printer.push(Separators.closingArguments);
	} else
	    visit(n.f2);
    }


    //
    // f0 -> Expression()
    // f1 -> ( "," Expression() )*
    //
    public void visit(ArgumentList n) {
	visit(n.f0);
	visit(n.f1, Separators.empty, Separators.empty,
	      Separators.betweenArguments);
    }

    //
    // f0 -> "new" PrimitiveType() ArrayDimsAndInits()
    //       | "new" Name() ( ArrayDimsAndInits()
    //                        | Arguments() [ ClassBody() ] )
    //
    public void visit(AllocationExpression n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;

	seq.elementAt(0).accept(this);
	printer.push(Separators.allocation);
	seq.elementAt(1).accept(this);
	if (n.f0.which == 0)
	    seq.elementAt(2).accept(this);
	else {
	    NodeChoice nodeChoice = (NodeChoice) seq.elementAt(2);
	    if (nodeChoice.which == 0)
		nodeChoice.choice.accept(this);
	    else {
		seq = (NodeSequence) nodeChoice.choice;
		seq.elementAt(0).accept(this);
		if (((NodeOptional) seq.elementAt(1)).present()) {
		    printer.push(Separators.control);
		    updateAnonymousClassLevel(+1);
		    seq.elementAt(1).accept(this);
		    updateAnonymousClassLevel(-1);
		}
	    }
	}
    }

    
    //
    // f0 -> ( "[" Expression() "]" )+ ( "[" "]" )*
    //       | ( "[" "]" )+ ArrayInitializer()
    //
    public void visit(ArrayDimsAndInits n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	if (n.f0.which == 0)
	    seq.accept(this);
	else {
	    seq.elementAt(0).accept(this);
	    printer.push(Separators.declarator);
	    seq.elementAt(1).accept(this);
	}
    }


    //
    // f0 -> LabeledStatement()
    //       | Block()
    //       | EmptyStatement()
    //       | StatementExpression() ";"
    //       | SwitchStatement()
    //       | IfStatement()
    //       | WhileStatement()
    //       | DoStatement()
    //       | ForStatement()
    //       | BreakStatement()
    //       | ContinueStatement()
    //       | ReturnStatement()
    //       | ThrowStatement()
    //       | SynchronizedStatement()
    //       | TryStatement()
    //
    public void visit(Statement n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> < IDENTIFIER >
    // f1 -> ":"
    // f2 -> Statement()
    //
    public void visit(LabeledStatement n) {
	visit(n.f0);
	printer.push(Separators.beforeLabel);
	visit(n.f1);
	printer.push(Separators.afterLabel);
	visit(n.f2);
    }

    //
    // f0 -> "{"
    // f1 -> ( BlockStatement() )*
    // f2 -> "}"
    //
    public void visit(Block n) {
	visit(n.f0);
	if ( n.f1.present() ) {
	    printer.push(Separators.newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f1, Separators.newLine, Separators.newLine);
	    spc.visitorUpdate(-1);
	}
	visit(n.f2);
    }

    //
    // f0 -> LocalVariableDeclaration() ";"
    //       | Statement()
    //       | UnmodifiedClassDeclaration()
    //       | UnmodifiedInterfaceDeclaration()
   //
    public void visit(BlockStatement n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> [ "final" ]
    // f1 -> Type()
    // f2 -> VariableDeclarator()
    // f3 -> ( "," VariableDeclarator() )*
    //
    public void visit(LocalVariableDeclaration n) {
	visit(n.f0, Separators.declarator);
	visit(n.f1);
	printer.push(Separators.declarator);
	visit(n.f2);
	visit(n.f3, Separators.empty, Separators.empty, Separators.declarator);
    }

    //
    // f0 -> ";"
    //
    public void visit(EmptyStatement n) {
	visit(n.f0);
    }

    //
    // f0 -> PreIncrementExpression()
    //       | PreDecrementExpression()
    //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
    //
    public void visit(StatementExpression n) {
	if (n.f0.which == 2) {
	    NodeSequence seq = (NodeSequence) n.f0.choice;
	    NodeOptional nodeOptional = (NodeOptional) seq.elementAt(1);
	    if (nodeOptional.present()
		   && ((NodeChoice) nodeOptional.node).which == 2) {
		seq.elementAt(0).accept(this);
		printer.push(Separators.beforeAssignment);
		visit((NodeSequence) ((NodeChoice) nodeOptional.node).choice,
		      Separators.afterAssignment);
		return;
	    }
	}
	n.f0.choice.accept(this);
    }

    //
    // f0 -> "switch"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> "{"
    // f5 -> ( SwitchLabel() ( BlockStatement() )* )*
    // f6 -> "}"
    //
    /** A case label that is not preceded by a break, a return or a continue
     *  statement is marked by a "// fall through" comment.
     */
    public void visit(SwitchStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
	printer.push(Separators.control);
	visit(n.f4);
	printer.push(Separators.newLine);
	spc.visitorUpdate(+1);
	boolean fallThroughShouldBePrinted = false;
	
	for (Enumeration e = n.f5.elements(); e.hasMoreElements();) {
	    NodeSequence seq = (NodeSequence) e.nextElement();

	    if (fallThroughShouldBePrinted &&
		fallThroughNotYetThere((SwitchLabel) seq.elementAt(0))) {
		    printer.push("// fall through");
		    printer.push(Separators.newLine);
		    fallThroughShouldBePrinted = false;
	    }
	    
	    seq.elementAt(0).accept(this);
	    printer.push(Separators.afterCaseLabel);
	    NodeListOptional list = (NodeListOptional) seq.elementAt(1);
	    if (list.present()) {
		printer.push(Separators.newLine);
		spc.visitorUpdate(+1);
		visit(list, Separators.newLine, Separators.newLine);

		// Check if a "fall through"-comment is needed.
		if (e.hasMoreElements()) {
		    NodeChoice nodeChoice = (NodeChoice)
			((BlockStatement) list.elementAt(list.size() - 1)).f0;
		    if (nodeChoice.which == 1) {
			switch (((NodeChoice)
				 ((Statement) nodeChoice.choice).f0) .which) {
			case 9: case 10: case 11: break;
			default:
			    fallThroughShouldBePrinted = true;
			}
		    }
		}
		spc.visitorUpdate(-1);
	    } else if (!e.hasMoreElements()) {
		printer.push(Separators.newLine);
	    }
	}
	spc.visitorUpdate(-1);
	visit(n.f6);
    }


    /** A case label that is not preceded by a break, a return or a continue
     *  statement is marked by a "// fall through" comment.
     *  This method checks if a similar comment already exists.
     */
    public boolean fallThroughNotYetThere(SwitchLabel n) {
	Token first = (Token) ((NodeSequence) n.f0.choice).elementAt(0);
	for (Enumeration e = new PrecedingSpecialTokensEnumeration(first);
	        e.hasMoreElements();) {
	    String comment = e.nextElement().toString().toLowerCase();
	    if (comment.indexOf("fall") != -1
		    && comment.indexOf("through") != -1)
		return false;
	}
	return true;
    }
    
    
    //
    // f0 -> "case" Expression() ":"
    //       | "default" ":"
    //
    public void visit(SwitchLabel n) {
	NodeSequence seq = (NodeSequence) n.f0.choice;
	seq.elementAt(0).accept(this);
	if (n.f0.which == 0) {
	    printer.push(Separators.declarator);
	    seq.elementAt(1).accept(this);
	    printer.push(Separators.beforeCaseLabel);
	    seq.elementAt(2).accept(this);
	} else
	    seq.elementAt(1).accept(this);
    }

    //
    // f0 -> "if"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Statement()
    // f5 -> [ "else" Statement() ]
    //
    public void visit(IfStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);

	boolean isIfBlock = (n.f4.f0.which == 1);
	if (!isIfBlock) {    // only indent if it's not a Block
	    spc.visitorUpdate(+1);
	    printer.push(Separators.newLine);
	} else
	    printer.push(Separators.control);
	visit(n.f4);
	if (!isIfBlock)
	    spc.visitorUpdate(-1);

	if ( n.f5.present() ) {
	    if (!isIfBlock)
		printer.push(Separators.newLine);
	    else 
		printer.push(Separators.control);

	    int kindOfElseBlock = ((Statement)
				   ((NodeSequence) n.f5.node).elementAt(1))
		                  .f0.which;
	  
	    if (kindOfElseBlock != 1 && kindOfElseBlock != 5) {
		NodeSequence seq = (NodeSequence) n.f5.node;
		seq.elementAt(0).accept(this);
		printer.push(Separators.newLine);
		spc.visitorUpdate(+1);
		seq.elementAt(1).accept(this);
		spc.visitorUpdate(-1);
	    } else
		visit((NodeSequence) n.f5.node, Separators.control);
	}
    }

    //
    // f0 -> "while"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Statement()
    //
    public void visit(WhileStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);

	if (n.f4.f0.which != 1) {
	    printer.push(Separators.newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f4);
	    spc.visitorUpdate(-1);
	} else {
	    printer.push(Separators.control);
	    visit(n.f4);
	}
    }

    //
    // f0 -> "do"
    // f1 -> Statement()
    // f2 -> "while"
    // f3 -> "("
    // f4 -> Expression()
    // f5 -> ")"
    // f6 -> ";"
    //
    public void visit(DoStatement n) {
	visit(n.f0);

	if (n.f1.f0.which != 1) {
	    printer.push(Separators.newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f1);
	    spc.visitorUpdate(-1);
	    printer.push(Separators.newLine);
	} else {
	    printer.push(Separators.control);
	    visit(n.f1);
	    printer.push(Separators.control);
	}
	visit(n.f2);
	printer.push(Separators.control);
	visit(n.f3);
	visit(n.f4);
	visit(n.f5);
	visit(n.f6);
    }

    //
    // f0 -> "for"
    // f1 -> "("
    // f2 -> [ ForInit() ]
    // f3 -> ";"
    // f4 -> [ Expression() ]
    // f5 -> ";"
    // f6 -> [ ForUpdate() ]
    // f7 -> ")"
    // f8 -> Statement()
    //
    public void visit(ForStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	printer.push(Separators.openingArguments);
	visit(n.f2);
	visit(n.f3);
	if (n.f4.present()) {
	    printer.push(Separators.betweenArguments);
	    visit(n.f4);
	}
	visit(n.f5);
	if (n.f6.present()) {
	    printer.push(Separators.betweenArguments);
	    visit(n.f6);
	}
	visit(n.f7);
	printer.push(Separators.closingArguments);
      
	if ( n.f8.f0.which != 1 ) {
	    printer.push(Separators.newLine);
	    spc.visitorUpdate(+1);
	    visit(n.f8);
	    spc.visitorUpdate(-1);
	} else {
	    printer.push(Separators.control);
	    visit(n.f8); 
	}
    }

    //
    // f0 -> LocalVariableDeclaration()
    //       | StatementExpressionList()
    //
    public void visit(ForInit n) {
	n.f0.choice.accept(this);
    }

    //
    // f0 -> StatementExpression()
    // f1 -> ( "," StatementExpression() )*
    //
    public void visit(StatementExpressionList n) {
	visit(n.f0);
	visit(n.f1, Separators.empty, Separators.empty,
	      Separators.betweenArguments);
    }

    //
    // f0 -> StatementExpressionList()
    //
    public void visit(ForUpdate n) {
	visit(n.f0);
    }

    //
    // f0 -> "break"
    // f1 -> [ < IDENTIFIER >  ]
    // f2 -> ";"
    //
    public void visit(BreakStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.control);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "continue"
    // f1 -> [ < IDENTIFIER >  ]
    // f2 -> ";"
    //
    public void visit(ContinueStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.control);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "return"
    // f1 -> [ Expression() ]
    // f2 -> ";"
    //
    public void visit(ReturnStatement n) {
	visit(n.f0);
	if (n.f1.present()) {
	    printer.push(Separators.control);
	    visit(n.f1);
	}
	visit(n.f2);
    }

    //
    // f0 -> "throw"
    // f1 -> Expression()
    // f2 -> ";"
    //
    public void visit(ThrowStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	visit(n.f2);
    }

    //
    // f0 -> "synchronized"
    // f1 -> "("
    // f2 -> Expression()
    // f3 -> ")"
    // f4 -> Block()
    //
    public void visit(SynchronizedStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	visit(n.f2);
	visit(n.f3);
	printer.push(Separators.control);
	visit(n.f4);
    }

    //
    // f0 -> "try"
    // f1 -> Block()
    // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
    // f3 -> [ "finally" Block() ]
    //
    public void visit(TryStatement n) {
	visit(n.f0);
	printer.push(Separators.control);
	visit(n.f1);
	for ( Enumeration e = n.f2.elements(); e.hasMoreElements(); ) {
	    NodeSequence seq = (NodeSequence) e.nextElement();
	    printer.push(Separators.control);
	    seq.elementAt(0).accept(this);
	    printer.push(Separators.control);
	    seq.elementAt(1).accept(this);
	    seq.elementAt(2).accept(this);
	    seq.elementAt(3).accept(this);
	    printer.push(Separators.control);
	    seq.elementAt(4).accept(this);
	}

	if (n.f3.present()) {
	    printer.push(Separators.control);
	    visit(n.f3, Separators.empty, Separators.control);
	}
    }
}






/** PrintStack manages a sequence of outputs that may exceed 80 characters or
 *  another line length (specified in the constructor). If necessary, the
 *  optimal location of line breaks is found - based on the priority and type
 *  information of the Separator objects.
 *  <p>
 *  Indentation is managed by a Spacing object; broken lines are indented
 *  twice e.g. in order to make long conditions in if-statements more readable.
 */
class PrintStack {

    /** Destination of the output */
    protected final PrintWriter out;
    /** The maximal line length */
    protected final int lineLength;
    /** Indentation manager */
    protected final Spacing spc;
    /** The column where "//"-comments occuring inside or behind an
     *	expression should start
     */
    protected final int commentColumn;
    /** The blanks String for a "//"-comment indented to commentColumn
     *  which appears alone on its line
     */
    protected final String blanksToCommentColumn;

    
    /** The current column in the yet non-broken line */
    protected int currentColumn;

    /** The number of empty lines just before the current line.
     *  It's needed for reconciling the newLines in the original text (saved
     *  as special tokens, see the grammar) and the line separation standards
     *  used in the PrettyPrinter. The former is taken into account only if it
     *  is larger than the latter at the current location.
     *  <p>
     *  @see push
     */
    protected int numberOfNewLines = 0;
    
    /** Variable for output text that hasn't been followed by a Separator
     *  object.
     *  <p>
     *  @see push
     */
    protected String waitingOutput = null;
    
    /** The stack for saving output items consisting of an output text and a
     *  Separator object.
     *  <p>
     *  @see Item
     */
    protected Vector items = new Vector();

    /** For saving "//"-comments occuring inside or behind an expression */
    protected Vector singleLineComments = new Vector();

    /** The space between the stars and the text in formal comments */
    protected String formalCommentSeparatorFromStars;

    
    /** Initializes the final attributes. */
    public PrintStack(PrintWriter out, Spacing spc, int lineLength,
		      int commentColumn) {
	this.out = out;
	this.spc = spc;
	this.lineLength = lineLength;
	this.commentColumn = commentColumn;
	StringBuffer buffer = new StringBuffer();
	for (int i = 0; i < commentColumn; ++i)
	    buffer.append(" ");
	blanksToCommentColumn = buffer.toString();
    }

    /** Here commentColumn is set to lineLength/2 */
    public PrintStack(PrintWriter out, Spacing spc, int lineLength) {
	this(out, spc, lineLength, lineLength/2);
    }

    
    /** Pushes an output string onto the stack. As long as it isn't followed
     *  by a Separator, it is saved by the waitingOutput reference.
     *  If there is already a waitingOutput, an empty Separator - the default -
     *  is used.
     */
    public void push(String output) {
	if (waitingOutput != null)
	    push(Separators.empty);
	waitingOutput = output;
    }

    /** Pushes a Separator onto the stack. An Item object is composed of the
     *  waitingOutput and this Separator. The new endColumn in the non-broken
     *  line is computed and used for initialization of the item object.
     *  A non-existing waitingOutput object is interpreted as an empty string.
     *  Updates of the Spacing object by the visitor are notified whenever
     *  a first item in a line is pushed onto the stack.
     */
    public void push(Separator newSeparator) {
	if (waitingOutput == null)
	    waitingOutput = "";
	
	if (items.size() == 0) {
	    spc.notifyVisitorUpdates();
	    currentColumn = spc.getWidth();
	}
	
	if (newSeparator instanceof NewLineSeparator) {
	    currentColumn += waitingOutput.length() + 1; 
	    items.addElement(new Item(waitingOutput, Separators.empty,
				      currentColumn));
	    flush();
	    
	    NewLineSeparator sep = (NewLineSeparator) newSeparator;
	    int numberOfLinesToPrint = sep.numberOfLinesToPrint;
	    
	    // singleLineComments carry already one newLine character
	    if (singleLineComments.size() > 0 && numberOfLinesToPrint > 0)
		--numberOfLinesToPrint;
	    flushSingleLineComments();
	    
	    for (int i = 0; i < numberOfLinesToPrint; ++i)
		out.println();
	    numberOfNewLines = sep.numberOfNewLines;
	} else {
	    currentColumn += waitingOutput.length() + newSeparator.length();
	    items.addElement(new Item(waitingOutput, newSeparator,
				      currentColumn));
	}
	waitingOutput = null;
    }

    
    /** Handles output of all the special Tokens - comments and user-specified
     *  newLines. "//"-comments appearing inside an expression are saved in the
     *  singleLineComments vector and output as soon as a NewLineSeparator is
     *  pushed.
     *
     *  @see flushSingleLineComments
     */
    public void pushSpecialToken(Token t) {
	String text;
	boolean beforeAnyOtherOutput = (items.size() == 0);
	switch (t.kind) {
	case JavaParserConstants.NEW_LINE:
	    if (beforeAnyOtherOutput) {
		if (numberOfNewLines > 0)
		    numberOfNewLines--;
		else 
		    out.println();
	    } 
	    break;
        case JavaParserConstants.MULTI_LINE_COMMENT_FIRST_LINE:
	    if (!beforeAnyOtherOutput)
		push(Separators.newLine);
	    push(t.toString());
	    push(Separators.newLineAlreadyInText);
	    break;
	case JavaParserConstants.MULTI_LINE_COMMENT_INTERMEDIATE_LINE:
	    push(t.toString());
	    push(Separators.newLineAlreadyInText);
	    break;
        case JavaParserConstants.MULTI_LINE_COMMENT_FINAL_LINE:
	    push(t.toString());
	    push(Separators.newLine);
	    break;
	case JavaParserConstants.FORMAL_COMMENT_FIRST_LINE:
	    if (!beforeAnyOtherOutput)
		push(Separators.newLine);
	    text = t.toString().trim();
	    formalCommentSeparatorFromStars =
		(text.length() == 3) ? " " : "  ";
	    push(text);
	    push(Separators.newLine);
	    break;
	case JavaParserConstants.FORMAL_COMMENT_INTERMEDIATE_LINE:
	    text = t.toString().trim();
	    if (text.startsWith("*")) 
		push(" " + text);
	    else
	        push(" *" + formalCommentSeparatorFromStars + text);
	    push(Separators.newLine);
	    break;
        case JavaParserConstants.FORMAL_COMMENT_FINAL_LINE:
	    push(" " + t.toString().trim());
	    push(Separators.newLine);
	    break;
	case JavaParserConstants.SINGLE_LINE_COMMENT:
	    if (beforeAnyOtherOutput) {
		push(t.toString());
		push(Separators.newLineAlreadyInText);
		numberOfNewLines = 0;
	    } else 
		singleLineComments.addElement(t.toString());
	    break;
        case JavaParserConstants.C_COMMENT_ON_SINGLE_LINE:
	case JavaParserConstants.FORMAL_COMMENT_ON_SINGLE_LINE:
	    push(t.toString());
	    if (beforeAnyOtherOutput)
		push(Separators.newLine);
	    break;
	}
    }

    
    /** An inner class for managing an output substring followed by a Separator.
     *  It also saves the endColumn of this part of the non-broken line.
     */
    protected static class Item {
	public final String output;
	public final Separator separator;
	public final int endColumn;
	public Item(String output, Separator separator, int endColumn) {
	    this.output = output;
	    this.separator = separator;
	    this.endColumn = endColumn;
	}
	public String toString() {
	    return output + separator;
	}
    }


    /** Prints a range of printStack items between the indices begin and end,
     *  inclusively.
     */
    protected void printRangeAsLine(int begin, int end) {
	out.print(spc.spc);
	for (int i = begin; i < end; ++i) {
	    out.print(items.elementAt(i).toString());
	}
	out.print(((Item) items.elementAt(end)).output);
	out.flush();
    }
	

    /** Flushes the print stack, because a NewLineSeparator has just been
     *  pushed onto the stack. <br>
     *  If the line doesn't exceed the lineLength, just print out all items
     *  (using printRangeAsLine). Otherwise determine the optimal Separators
     *  and break the line at these Separators. <br>
     *  All but the first line are indented by two amounts of the Spacing
     *  object.
     *  Ultimately, all the elements are removed from the stack.
     *
     *  @see printRangeAsLine, findMaximalSeparator
     */
    protected void flush() {
	int length = items.size();
	if (currentColumn <= lineLength || length == 1) {
	    printRangeAsLine(0, length - 1);
        } else {
	    int begin = 0;
	    
	    // Print the first line
	    int end = findMaximalSeparator(begin);
	    printRangeAsLine(begin, end);
	    begin = end + 1;

	    // Indent two amounts and print the other lines.
	    spc.printerUpdate(+2);
	    while (begin < length) {
		end = findMaximalSeparator(begin);
		out.println();
		printRangeAsLine(begin, end);
		begin = end + 1;
	    }
	    spc.printerUpdate(-2);
	}
	items.removeAllElements();
    }


    /** Flushes all the singleLineComments that were encountered inside and at
     *  the end of the current expression.
     */
    public void flushSingleLineComments() {
	if (singleLineComments.size() == 0)
	    return;
	int i;
	String firstSingleLineComment =
	    (String) singleLineComments.elementAt(0);
	if (currentColumn < commentColumn) {
	    for (i = 0; i <= commentColumn - currentColumn; ++i)
		out.print(" ");
	} else {
	    if (firstSingleLineComment.length() + currentColumn - 1
		    < lineLength)
		out.print(" ");
	    else {
		out.println();
		out.print(blanksToCommentColumn);
	    }
	}
	out.print(firstSingleLineComment);
	for (i = 1; i < singleLineComments.size(); ++i) {
	    out.print(blanksToCommentColumn);
	    out.print((String) singleLineComments.elementAt(i));
	}
	singleLineComments.removeAllElements();
    }

    
    /** Finds the optimal Separator in the print stack after the index "begin".
     *  A sequence of SpecialSeparators occuring at the beginning of the stack
     *  is treated first. They should stay together and may only be separated
     *  if they are larger than lineLength. <br>
     *  NormalSeparators and non-initial SpecialSeparators are judged by their
     *  priority. The last one carrying the maximal priority is chosen. <br>
     *  Nested expressions have their priorities lowered by the amount
     *  Separators.MAX_PRIORITY.
     */
    protected int findMaximalSeparator(int begin) {
	int length = items.size();

	// currentItem.endColumn - columnOffset is the endColumn in the broken
	// line, while currentItem.endColumn is the endColumn in the non-broken
	// line. columnOffset subtracts the indentation and the column of the
        // Item that has been output most recently.
	int columnOffset = 0;
	if (begin > 0) {
	    columnOffset = ((Item) items.elementAt(begin - 1)).endColumn
		- spc.getWidth();
	}

	// In case this is the last part of a broken line:
	if ((currentColumn =
	          ((Item) items.elementAt(length - 1)).endColumn - columnOffset)
   	          <= lineLength) {
	    return length - 1;
	}
	
	int maxPriority = -100000;
	int maxPriorityIndex = begin;
	boolean yetOnlySpecialSeparators = true;
	int priorityOffset = 0;
	
	for (int i = begin; i < length; ++i) {
	    Item currentItem = (Item) items.elementAt(i);
	    
	    // In case this Item belongs into the next line:
	    if (currentItem.endColumn - columnOffset > lineLength)
		break;

	    int priority = currentItem.separator.priority + priorityOffset;
	    
	    // Adjustment of priority for nested expressions
	    if (currentItem.separator instanceof SpecialSeparator) {
		if (currentItem.separator == Separators.openingArguments)
		    priorityOffset -= Separators.MAX_PRIORITY;
		else if (currentItem.separator == Separators.closingArguments) {
		    priorityOffset += Separators.MAX_PRIORITY;
		    priority += Separators.MAX_PRIORITY;
		} else if (currentItem.separator == Separators.betweenArguments)
		    priority += Separators.MAX_PRIORITY;
	    } else if (yetOnlySpecialSeparators) {
		// The sequence of SpecialSeparators ends here:
		yetOnlySpecialSeparators = false;
	    }

	    // Don't break the line at items within the first three INDENT_AMTs
	    // of the line.
	    if (begin == 0 &&
		   currentItem.endColumn - spc.getWidth()
		   <= 3 * spc.INDENT_AMT)
		continue;

	    // Update the optimum.
	    // Use >= because we want to find the last maximal one.
	    if (priority >= maxPriority) {
		maxPriority = priority;
		maxPriorityIndex = i;
	    }

	    // If we're inside a sequence of SpecialSeparators, ignore the
	    // former maxPriorityIndices and consider the current Separator
	    // as the optimal one. This keeps the SpecialSeparators together.
	    // Don't treat empty Separators (they are SpecialSeparators) this
	    // way.
	    if (yetOnlySpecialSeparators
		&& currentItem.separator != Separators.empty)
		maxPriorityIndex = i;

	}
	return maxPriorityIndex;
    }

}


/** A version of PrintStack used for debugging purposes */
class PrintStackDebugged extends PrintStack {
    
    public PrintStackDebugged(PrintWriter out, Spacing spc, int lineLength) {
	super(out, spc, lineLength);
    }
    
    /** Displays the Separator priorities instead of the Separators */
    protected void printRangeAsLine(int begin, int end) {
	out.print(spc.spc);
	for (int i = begin; i <= end; ++i) {
	    out.print(((Item) items.elementAt(i)).output + "_" +
		      ((Item) items.elementAt(i)).separator.priority + "_");
	}
    }
    
}


/** Separators saves all the Separator constants used in the PrettyPrinter. */
interface Separators {
    public static final SpecialSeparator empty = new SpecialSeparator("", 0);

    public static final NewLineSeparator newLine =
	new NewLineSeparator(1);
    public static final NewLineSeparator doubleNewLine =
	new NewLineSeparator(2);
    public static final NewLineSeparator tripleNewLine =
	new NewLineSeparator(3);
    public static final NewLineSeparator newLineAlreadyInText =
	new NewLineSeparator(0, 1);
    
    public static final NormalSeparator cast =
	new NormalSeparator(" ", 1);
    public static final NormalSeparator allocation =
	new NormalSeparator(" ", 1);
    public static final SpecialSeparator openingArguments =
	new SpecialSeparator("", 2);
    public static final SpecialSeparator betweenArguments =
	new SpecialSeparator(" ", 2);
    public static final SpecialSeparator declarator =
	new SpecialSeparator(" ", 3);
    public static final SpecialSeparator declaratorEmpty =
	new SpecialSeparator("", 3);
    public static final NormalSeparator subDeclarator =
	new NormalSeparator(" ", 2);
    public static final SpecialSeparator closingArguments =
	new SpecialSeparator("", 0);
    public static final SpecialSeparator beforeDot =
	new SpecialSeparator("", 3);
    public static final NormalSeparator beforeMultiplicative =
	new NormalSeparator(" ", 3);
    public static final NormalSeparator afterMultiplicative =
	new NormalSeparator(" ", 4);
    public static final SpecialSeparator beforeAdditive =
	new SpecialSeparator(" ", 4);
    public static final SpecialSeparator afterAdditive =
	new SpecialSeparator(" ", 5);
    public static final NormalSeparator beforeShift =
	new NormalSeparator(" ", 5);
    public static final NormalSeparator afterShift =
	new NormalSeparator(" ", 6);
    public static final NormalSeparator beforeRelational =
	new NormalSeparator(" ", 6);
    public static final NormalSeparator afterRelational =
	new NormalSeparator(" ", 7);
    public static final NormalSeparator beforeInstanceOf =
	new NormalSeparator(" ", 7);
    public static final NormalSeparator afterInstanceOf =
	new NormalSeparator(" ", 8);
    public static final NormalSeparator beforeEquality =
	new NormalSeparator(" ", 8);
    public static final NormalSeparator afterEquality =
	new NormalSeparator(" ", 9);
    public static final NormalSeparator beforeAnd =
	new NormalSeparator(" ", 9);
    public static final NormalSeparator afterAnd =
	new NormalSeparator(" ", 10);
    public static final NormalSeparator beforeExclusiveOr =
	new NormalSeparator(" ", 10);
    public static final NormalSeparator afterExclusiveOr =
	new NormalSeparator(" ", 11);
    public static final NormalSeparator beforeInclusiveOr =
	new NormalSeparator(" ", 11);
    public static final NormalSeparator afterInclusiveOr =
	new NormalSeparator(" ", 12);
    public static final NormalSeparator beforeConditionalAnd =
	new NormalSeparator(" ", 12);
    public static final NormalSeparator afterConditionalAnd =
	new NormalSeparator(" ", 13);
    public static final NormalSeparator beforeConditionalOr =
	new NormalSeparator(" ", 13);
    public static final NormalSeparator afterConditionalOr =
	new NormalSeparator(" ", 14);
    public static final NormalSeparator conditional =
	new NormalSeparator(" ", 14);
    public static final NormalSeparator beforeLabel =
	new NormalSeparator("", 0);
    public static final NormalSeparator afterLabel =
	new NormalSeparator(" ", 17);
    public static final NormalSeparator beforeCaseLabel =
	new NormalSeparator("", 0);
    public static final NormalSeparator afterCaseLabel =
	new NormalSeparator(" ", 17);
    public static final SpecialSeparator beforeAssignment =
	new SpecialSeparator(" ", 17);
    public static final SpecialSeparator afterAssignment =
	new SpecialSeparator(" ", 17);
    public static final SpecialSeparator control =
	new SpecialSeparator(" ", 17);

    public static final int MAX_PRIORITY = 17;
}


/** A Separator separates the output items (Tokens or comments) of the
 *  prettyprinter in the Printstack. It is a potential location of a line
 *  break. Each Separator has a certain priority to be used for breaking
 *  a large line into parts.
 */
abstract class Separator {
    protected final String text;
    protected final int priority;
    
    public int getPriority() { return priority; }
    public String toString() { return text; }
    public int length() { return text.length(); }
    
    public Separator(String text, int priority) {
	this.text = text;
	this.priority = priority;
    }

}


/** NormalSeparator are sorted by priority. The last NormalSeparator inside
 *  the space of one ouput line with maximal priority is used for breaking
 *  the line.
 */
class NormalSeparator extends Separator {
    public NormalSeparator(String text, int priority) {
	super(text, priority);
    }
}
    

/** A NewLineSeparator begins implicitly with a new line, which is omitted in
 *  the text part. There may be more new line characters inside.
 */ 
class NewLineSeparator extends Separator {
    protected final int numberOfNewLines;
    protected final int numberOfLinesToPrint;
    public NewLineSeparator(int numberOfLinesToPrint, int numberOfNewLines) {
	super("", 0);
	this.numberOfNewLines = numberOfNewLines;
	this.numberOfLinesToPrint = numberOfLinesToPrint;
    }
    public NewLineSeparator(int numberOfLinesToPrint) {
	this(numberOfLinesToPrint, numberOfLinesToPrint);
    }
}

/** If a line begins with a sequence of items separated only by
 *  SpecialSeparators these items are first enumerated without consideration
 *  to priority. Only after the first non-SpecialSeparator, priority is taken
 *  into account and the remaining SpecialSeparators are treated just like
 *  NormalSeparators.
 */
class SpecialSeparator extends Separator {
    public SpecialSeparator(String text, int priority) {
	super(text, priority);
    }
}



/** Spacing manages the indentation of the code. Update requests from the
 *  PrettyPrinter visitor and from the PrintStack are treated differently,
 *  since a visitor request may occur before the NewLineSeparator has been
 *  pushed onto the PrintStack. This could corrupt the printer's own
 *  indentation, which only starts after the NewLineSeparator has been
 *  received. <br>
 *  The visitor requests are only executed after the printer calls
 *  notifyVisitorUpdates().
 */
class Spacing {
    final int INDENT_AMT;

    String spc = "";
    int indentLevel = 0;

    /** The width (in characters) of the indentation */
    int width = 0;

    /** Accumulates the visitor updates that haven't been processed yet */
    int preparedUpdate = 0;
    
    public Spacing(int indentAmt) {
	INDENT_AMT = indentAmt;
    }

    public String toString() {
	return spc;
    }

    public int getWidth() {
	return width;
    }


    /** Receives an update request from the visitor. It isn't executed until
     *  notifyVisitorUpdates is called
     */
    public void visitorUpdate(int numIndentLvls) {
	preparedUpdate += numIndentLvls;
    }

    /** Executes the updated requests from the visitor. */
    public void notifyVisitorUpdates() {
	if (preparedUpdate != 0) {
	    printerUpdate(preparedUpdate);
	    preparedUpdate = 0;
	}
    }

    /** Receives and executes an update request from the printer. */
    public void printerUpdate(int numIndentLvls) {
	indentLevel += numIndentLvls;
        int widthIncrement = numIndentLvls * INDENT_AMT;
	width += widthIncrement;
	if (numIndentLvls < 0) {
	    spc = spc.substring(-widthIncrement);
	} else if (numIndentLvls > 0) {
	    StringBuffer buf = new StringBuffer(spc);
	    for (int i = 0; i < widthIncrement; ++i)
		buf.append(" ");
	    spc = buf.toString();
	}
    }
}


/** An Enumeration of the current Token's specialTokens appearing before this
 *  Token. The first "//"-comment is assigned to the previous Token (except
 *  for the case that we are in the first line of input - then there is no
 *  previous Token). The same is true for c-comments appearing on the same line
 *  as the previous expression.
 */
class PrecedingSpecialTokensEnumeration implements Enumeration {
    protected Token currentSpecialToken;
    public PrecedingSpecialTokensEnumeration(Token t) {
	if (t.specialToken == null) {
	    currentSpecialToken = null;
	    return;
	}
	    
	Token traverser = t.specialToken;
	Token firstSpecialTokenBefore = null;
	while (traverser.specialToken != null) {
	    switch (traverser.specialToken.kind) {
	    case JavaParserConstants.SINGLE_LINE_COMMENT:
		firstSpecialTokenBefore = traverser;
		break;
	    case JavaParserConstants.MULTI_LINE_COMMENT_FIRST_LINE:
	    case JavaParserConstants.FORMAL_COMMENT_FIRST_LINE:
		firstSpecialTokenBefore = traverser.specialToken;
		break;
	    case JavaParserConstants.NEW_LINE:
		firstSpecialTokenBefore = traverser.specialToken;
		break;
	    }
	    traverser = traverser.specialToken;
	}
	if (traverser.beginLine == 1) {
	    currentSpecialToken = traverser;
	} else
	    currentSpecialToken = firstSpecialTokenBefore;
    }

    public boolean hasMoreElements() {
	return currentSpecialToken != null;
    }

    public Object nextElement() {
	Token result = currentSpecialToken;
	currentSpecialToken = currentSpecialToken.next;
	return result;
    }
}


/** An Enumeration of the current Token's specialTokens appearing behind this
 *  Token. The first "//"-comment is assigned to the this Token. The same is
 *  true for c-comments appearing on the same line as this expression.
 */
class FollowingSpecialTokensEnumeration implements Enumeration {
    protected Token currentSpecialToken;
    protected Token stopAtSpecialToken = null;
    public FollowingSpecialTokensEnumeration(Token t) {
	if (t.next == null || t.next.specialToken == null) {
	    currentSpecialToken = null;
	    return;
	}
	// Let traverser go to the first specialToken of the next Token
	Token traverser = t.next.specialToken;
	while (traverser.specialToken != null)
	    traverser = traverser.specialToken;
	currentSpecialToken = traverser;

	// Let traverser go to the last one of next's specialTokens
	// that belongs to this Token
	while (traverser != null) {
	    switch (traverser.kind) {
	    case JavaParserConstants.SINGLE_LINE_COMMENT:
		stopAtSpecialToken = traverser.next;
		return;
	    case JavaParserConstants.NEW_LINE:
	    case JavaParserConstants.MULTI_LINE_COMMENT_FIRST_LINE:
	    case JavaParserConstants.FORMAL_COMMENT_FIRST_LINE:
		stopAtSpecialToken = traverser;
		return;
	    }
	    traverser = traverser.next;
	}
    }

    public boolean hasMoreElements() {
	return (currentSpecialToken != stopAtSpecialToken);
    }

    public Object nextElement() {
	Token result = currentSpecialToken;
	currentSpecialToken = currentSpecialToken.next;
	return result;
    }
}


